from pathlib import Path
from .cell_annotation_schema import *

THIS_PATH = Path(__file__).parent

SCHEMA_DIRECTORY = THIS_PATH.parent / "schema"
MAIN_SCHEMA_PATH = SCHEMA_DIRECTORY / "cell_annotation_schema_default.yaml"



# This snippet is from the rdflib_dumper.py file in the linkml_runtime/dumpers/rdflib_dumper.py
# Script adds handling for the Any class to be serialised as a json dump
# Please check `edit_start` and `edit_end` comments to see the changes

import logging
import urllib
from abc import abstractmethod
from typing import Optional, Any, Dict, Union
from pydantic import BaseModel

from curies import Converter
from rdflib import Graph, URIRef, XSD
from rdflib.term import Node, BNode, Literal
from rdflib.namespace import RDF


from linkml_runtime.dumpers.dumper_root import Dumper
from linkml_runtime.linkml_model import SlotDefinition
from linkml_runtime.utils.schemaview import SchemaView, ElementName, PermissibleValue, PermissibleValueText
from linkml_runtime.utils.yamlutils import YAMLRoot


class RDFLibDumper(Dumper):
    """
    Dumps from elements (instances of a LinkML model) to an rdflib Graph

    Note: this should be used in place of rdf_loader for now

    This requires a SchemaView object

    """
    def as_rdf_graph(
        self,
        element: Union[BaseModel, YAMLRoot],
        schemaview: SchemaView,
        prefix_map: Union[Dict[str, str], Converter, None] = None,
    ) -> Graph:
        """
        Dumps from element to an rdflib Graph,
        following a schema

        :param element: element to represent in RDF
        :param schemaview:
        :param prefix_map:
        :return:
        """
        g = Graph()
        if isinstance(prefix_map, Converter):
            # TODO replace with `prefix_map = prefix_map.bimap` after making minimum requirement on python 3.8
            prefix_map = {record.prefix: record.uri_prefix for record in prefix_map.records}
        logging.debug(f'PREFIXMAP={prefix_map}')
        if prefix_map:
            for k, v in prefix_map.items():
                if k == "@base":
                    schemaview.namespaces()._base = v
                else:
                    schemaview.namespaces()[k] = v
                    g.namespace_manager.bind(k, URIRef(v))
            for prefix in schemaview.namespaces():
                g.bind(prefix, URIRef(schemaview.namespaces()[prefix]))
        else:
            for prefix in schemaview.namespaces():
                g.bind(prefix, URIRef(schemaview.namespaces()[prefix]))
        # user can pass in base in prefixmap using '_base'. This gets set
        # in namespaces as a plain dict assignment - explicitly call the setter
        # to set the underlying "@base"
        if "_base" in schemaview.namespaces():
            schemaview.namespaces()._base = schemaview.namespaces()["_base"]
            g.base = schemaview.namespaces()._base
        if schemaview.namespaces()._base:
            g.base = schemaview.namespaces()._base
        self.inject_triples(element, schemaview, g)
        return g

    def inject_triples(self, element: Any, schemaview: SchemaView, graph: Graph, target_type: ElementName = None) -> Node:
        """
        Inject triples from conversion of element into a Graph

        :param element: element to represent in RDF
        :param schemaview:
        :param graph:
        :param target_type:
        :return: root node as rdflib URIRef, BNode, or Literal
        """
        namespaces = schemaview.namespaces()
        slot_name_map = schemaview.slot_name_mappings()
        logging.debug(f'CONVERT: {element} // {type(element)} // {target_type}')
        if target_type in schemaview.all_enums():
            if isinstance(element, PermissibleValueText):
                e = schemaview.get_enum(target_type)
                element = e.permissible_values[element]
            else:
                element = element.code
            element: PermissibleValue
            if element.meaning is not None:
                return URIRef(schemaview.expand_curie(element.meaning))
            else:
                return Literal(element.text)
        if target_type in schemaview.all_types():
            t = schemaview.get_type(target_type)
            dt_uri = t.uri
            if dt_uri:
                if dt_uri == 'rdfs:Resource':
                    return URIRef(schemaview.expand_curie(element))
                elif dt_uri == 'xsd:string':
                    return Literal(element)
                else:
                    if "xsd" not in namespaces:
                        namespaces["xsd"] = XSD
                    return Literal(element, datatype=namespaces.uri_for(dt_uri))
            else:
                logging.warning(f'No datatype specified for : {t.name}, using plain Literal')
                return Literal(element)
        #     edit_start
        if target_type == "Any":
            if isinstance(element, str):
                return Literal(element)
            elif isinstance(element, dict):
                import json
                return Literal(json.dumps(element))
        #     edit_end
        element_vars = {k: v for k, v in vars(element).items() if not k.startswith('_')}
        if len(element_vars) == 0:
            id_slot = schemaview.get_identifier_slot(target_type)
            return self._as_uri(element, id_slot, schemaview)
            #return URIRef(schemaview.expand_curie(str(element)))
        element_type = type(element)
        cn = element_type.class_name
        id_slot = schemaview.get_identifier_slot(cn)
        if id_slot is not None:
            element_id = getattr(element, id_slot.name)
            element_uri = self._as_uri(element_id, id_slot, schemaview)
        else:
            element_uri = BNode()
        type_added = False
        for k, v_or_list in element_vars.items():
            if isinstance(v_or_list, list):
                vs = v_or_list
            elif isinstance(v_or_list, dict):
                vs = v_or_list.values()
            else:
                vs = [v_or_list]
            for v in vs:
                if v is None:
                    continue
                if k in slot_name_map:
                    k = slot_name_map[k].name
                else:
                    logging.error(f'Slot {k} not in name map')
                slot = schemaview.induced_slot(k, cn)
                if not slot.identifier:
                    slot_uri = URIRef(schemaview.get_uri(slot, expand=True))
                    v_node = self.inject_triples(v, schemaview, graph, slot.range)
                    graph.add((element_uri, slot_uri, v_node))
                    if slot.designates_type:
                        type_added = True
        if not type_added:
            graph.add((element_uri, RDF.type, URIRef(schemaview.get_uri(cn, expand=True))))
        return element_uri

    def dump(
        self,
        element: Union[BaseModel, YAMLRoot],
        to_file: str,
        schemaview: SchemaView = None,
        fmt: str = 'turtle',
        prefix_map: Union[Dict[str, str], Converter, None] = None,
        **args,
    ) -> None:
        """
        Write element as rdf to to_file

        :param element: element to represent in RDF
        :param to_file:
        :param schemaview:
        :param fmt:
        :param prefix_map:
        :return:
        """
        super().dump(element, to_file, schemaview=schemaview, fmt=fmt, prefix_map=prefix_map)

    def dumps(
        self,
        element: Union[BaseModel, YAMLRoot],
        schemaview: SchemaView = None,
        fmt: Optional[str] = 'turtle',
        prefix_map: Union[Dict[str, str], Converter, None] = None,
    ) -> str:
        """
        Convert element into an RDF graph guided by the schema

        :param element:
        :param schemaview:
        :param fmt:
        :param prefix_map:
        :return: serialization of rdflib Graph containing element
        """
        return self.as_rdf_graph(element, schemaview, prefix_map=prefix_map).\
            serialize(format=fmt)

    def _as_uri(self, element_id: str, id_slot: Optional[SlotDefinition], schemaview: SchemaView) -> URIRef:
        if id_slot and schemaview.is_slot_percent_encoded(id_slot):
            return URIRef(urllib.parse.quote(element_id))
        else:
            return schemaview.namespaces().uri_for(element_id)




import requests

from typing import Union, Optional, List
from ruamel.yaml import YAML

from linkml_runtime.linkml_model import SchemaDefinition
from linkml_runtime.utils.schema_as_dict import schema_as_dict
from linkml_runtime.loaders import yaml_loader
from linkml_runtime.dumpers import json_dumper

from schema_automator.utils.schemautils import write_schema
from schema_automator.importers.jsonschema_import_engine import JsonSchemaImportEngine

from oaklib.utilities.subsets.value_set_expander import (
    ValueSetExpander,
    ValueSetConfiguration,
)


CAS_ROOT_CLASS = "Taxonomy"

CAS_NAMESPACE = "https://cellular-semantics.sanger.ac.uk/ontology/CAS"
CAS_ROOT_NS = "cell_annotation_schema"
DEFAULT_PREFIXES = {
    CAS_ROOT_NS: CAS_NAMESPACE + "/",
    "CAS": CAS_NAMESPACE + "/",
    "obo": "http://purl.obolibrary.org/obo/",
    "CL": "http://purl.obolibrary.org/obo/CL_",
    "PCL": "http://purl.obolibrary.org/obo/PCL_",
    "RO": "http://purl.obolibrary.org/obo/RO_",
    "skos": "http://www.w3.org/2004/02/skos/core#",
}


def decorate_linkml_schema(
    schema_obj: Union[dict, SchemaDefinition],
    output_path: Optional[str] = None,
) -> dict:
    """
    Adds additional properties to the LinkML schema necessary for Python conversion.
    Args:
        schema_obj (Union[dict, SchemaDefinition]): The LinkML schema object, either as a dictionary or a
        SchemaDefinition.
        output_path (Optional[str]): Path to the output schema file, if specified.
    Returns:
        dict: The schema decorated with additional properties for Python conversion.
    """
    if isinstance(schema_obj, SchemaDefinition):
        schema_obj = schema_as_dict(schema_obj)
    # schema_obj["id"] = CAS_NAMESPACE

    # this is a workaround for lack of annotation ids in the base schema
    if "Bican_Taxonomy" in schema_obj["classes"]:
        taxonomy_slots = list(schema_obj["classes"]["Taxonomy"]["slots"])
        taxonomy_slots.remove("annotations")
        schema_obj["classes"]["Taxonomy"]["slots"] = taxonomy_slots
        bican_annotation_slots = schema_obj["classes"]["Bican_Taxonomy"].get("slots", list())
        bican_annotation_slots.append("annotations")
        schema_obj["classes"]["Bican_Taxonomy"]["slots"] = bican_annotation_slots
    if "Cap_Taxonomy" in schema_obj["classes"]:
        taxonomy_slots = list(schema_obj["classes"]["Taxonomy"]["slots"])
        taxonomy_slots.remove("annotations")
        schema_obj["classes"]["Taxonomy"]["slots"] = taxonomy_slots
        cap_annotation_slots = schema_obj["classes"]["Cap_Taxonomy"].get("slots", list())
        cap_annotation_slots.append("annotations")
        schema_obj["classes"]["Cap_Taxonomy"]["slots"] = cap_annotation_slots

    for class_name in schema_obj["classes"]:
        if "attributes" in schema_obj["classes"][class_name]:
            clazz = schema_obj["classes"][class_name]
            del clazz["attributes"]

    if output_path:
        write_schema(schema_obj, output_path)

    return schema_obj


def decorate_linkml_ontology_schema(schema_obj: Union[dict, SchemaDefinition], ontology_iri: str, ontology_namespace: str, labelsets: Optional[List[str]] = None):
    """
    Decorates the LinkML schema with additional properties for OWL conversion.
    Args:
        schema_obj (Union[dict, SchemaDefinition]): The LinkML schema object, either as a dictionary or a
        SchemaDefinition.
        ontology_namespace (str): The namespace of the ontology (e.g., 'MTG').
        ontology_iri (str): The ontology IRI (e.g., 'https://purl.brain-bican.org/ontology/AIT_MTG/').
        labelsets (Optional[List[str]]): Labelsets used in the taxonomy, such as ['Cluster', 'Subclass', 'Class'].
    Returns:
        dict: The schema decorated with additional properties for OWL conversion.
    """
    if isinstance(schema_obj, SchemaDefinition):
        schema_obj = schema_as_dict(schema_obj)

    # these only required for ontology conversion
    ontology_namespace = ontology_namespace.upper()
    # prefixes = DEFAULT_PREFIXES.copy()
    # prefixes["linkml"] = "https://w3id.org/linkml/"
    prefixes = schema_obj["prefixes"]
    prefixes["_base"] = ontology_iri
    prefixes[ontology_namespace] = ontology_iri
    labelsets = labelsets or []
    for labelset in labelsets:
        prefixes[labelset] = ontology_iri + f"{labelset}#"
    schema_obj["prefixes"] = prefixes
    schema_obj["slots"]["id"] = {"identifier": True, "range": "uriorcurie"}
    schema_obj["classes"]["Labelset"]["slots"] = list(schema_obj["classes"]["Labelset"]["slots"]) + ["id"]
    schema_obj["classes"]["Taxonomy"]["slots"] = list(schema_obj["classes"]["Taxonomy"]["slots"]) + ["id"]
    # schema_obj["slots"]["cell_id"]["identifier"] = True   # TODO
    is_bican = "Bican_Taxonomy" in schema_obj["classes"]
    is_cap = "Cap_Taxonomy" in schema_obj["classes"]
    if "cell_set_accession" in schema_obj["slots"]:
        schema_obj["slots"]["cell_set_accession"]["identifier"] = True
    if "parent_cell_set_accession" in schema_obj["slots"]:
        schema_obj["slots"]["parent_cell_set_accession"]["range"] = "Bican_Annotation"
    if is_bican and "labelset" in schema_obj["slots"]:
        schema_obj["slots"]["labelset"]["range"] = "Bican_Labelset"

    schema_obj = decorate_linkml_schema(schema_obj)
    return schema_obj


def expand_schema(
    config: Optional[str],
    yaml_obj: dict,
    value_set_names: List[str],
    output_path: Optional[str] = None,
):
    """
    Dynamically expands the yaml_obj schema in-place using specified value set names.
    Args:
        config (Optional[str]): Path to the configuration file. If None, a default configuration is used.
        yaml_obj (dict): YAML schema object that will be expanded.
        value_set_names (List[str]): Names of the value sets to be included in the expansion.
        output_path (Optional[str]): Path where the expanded schema file will be saved, if specified.
    Note:
        Source code referenced from: https://github.com/INCATools/ontology-access-kit/blob/main/src/oaklib/utilities/subsets/value_set_expander.py
    """
    expander = ValueSetExpander()
    if config:
        expander.configuration = yaml_loader.load(
            config, target_class=ValueSetConfiguration
        )

    yaml = YAML()
    schema = yaml_loader.load(yaml_obj, target_class=SchemaDefinition)
    if value_set_names is None:
        value_set_names = list(schema.enums.keys())
    for value_set_name in value_set_names:
        if value_set_name not in schema.enums:
            raise ValueError(f"Unknown value set: {value_set_name}")
        value_set = schema.enums[value_set_name]
        pvs = list(expander.expand_value_set(value_set, schema=schema))
        yaml_obj["enums"][value_set_name]["permissible_values"] = {
            str(pv.text): json_dumper.to_dict(pv) for pv in pvs
        }
    if output_path:
        with open(output_path, "w", encoding="UTF-8") as file:
            yaml.dump(yaml_obj, stream=file)
    return yaml_obj



import os
import json
from pathlib import Path

from importlib import resources
from cell_annotation_schema import resource


class CurieToIriConverter:

    DEFAULT_JSONLD_FILE_NAME = 'prefixes.jsonld'

    def __init__(self, jsonld_file=DEFAULT_JSONLD_FILE_NAME):
        self.jsonld_file = jsonld_file
        self.prefix_mappings = self.load_prefix_mappings(self.jsonld_file)

    @staticmethod
    def load_prefix_mappings(jsonld_file_name):
        prefix_resource_file = resources.files(resource) / jsonld_file_name
        if os.path.exists(jsonld_file_name):
            # read from the path provided by the user
            with open(jsonld_file_name, 'r') as file:
                prefixes = json.load(file)
        elif os.path.exists(prefix_resource_file):
            # read from importlib.resources (resource) package
            prefixes = json.loads(Path(prefix_resource_file).read_text())
        else:
            # read from resource folder
            resource_dir = os.path.join(os.path.dirname(__file__), "./resource")
            prefixes_path = os.path.join(resource_dir, jsonld_file_name)
            prefixes = json.loads(Path(prefixes_path).read_text())
        return prefixes.get('@context', {})

    def curie_to_iri(self, curie):
        prefix, reference = curie.split(':', 1)
        if prefix in self.prefix_mappings:
            return self.prefix_mappings[prefix] + reference
        else:
            raise ValueError(f"Prefix '{prefix}' not found in prefix mappings.")



import click

from linkml_runtime.utils.schema_as_dict import schema_as_dict
from schema_automator.utils.schemautils import write_schema

from cell_annotation_schema.file_utils import read_schema
from cell_annotation_schema.ontology.schema import expand_schema


@click.group()
def cli():
    pass


@cli.command()
@click.option('-i', '--input', type=click.Path(exists=True), help='Input schema path.')
@click.option('-o', '--output', type=click.Path(), help='Output file path.')
def expand(input, output):
    schema_obj = read_schema(input)
    input_schema = schema_as_dict(schema_obj)
    expanded_schema = expand_schema(
        config=None, yaml_obj=input_schema, value_set_names=["CellTypeEnum"]
    )
    write_schema(expanded_schema, output)




if __name__ == '__main__':
    cli()


import rdflib

from pathlib import Path
from typing import Union, Optional, List

from cell_annotation_schema.file_utils import read_schema, get_json_from_file
from cell_annotation_schema.ontology.schema import (
    decorate_linkml_ontology_schema,
    expand_schema,
)
from cell_annotation_schema.ontology.data import dump_to_rdf, populate_ids


def export_to_rdf(
    cas_schema: Optional[Union[str, dict]],
    data: Union[str, dict],
    ontology_namespace: str,
    ontology_iri: str,
    output_path: str = None,
    validate: bool = True,
    include_cells: bool = True,
) -> rdflib.Graph:
    """
    Generates and returns an RDF graph from the provided data and CAS schema, with an option to write the RDF graph to a file.
    Args:
        cas_schema (Optional[Union[str, dict]]):
            Name of the CAS release (such as `base`, `cap`, `bican`), path to the CAS schema file, or CAS schema Yaml dict.
            If not provided, reads the `base` CAS schema.
        data (Union[str, dict]):
            The data JSON file path or JSON object dictionary.
        ontology_namespace (str):
            Ontology namespace (e.g., `MTG`).
        ontology_iri (str):
            Ontology IRI (e.g., `https://purl.brain-bican.org/ontology/AIT_MTG/`).
        labelsets (Optional[List[str]]):
            Labelsets used in the taxonomy, such as `["Cluster", "Subclass", "Class"]`.
        output_path (Optional[str]):
            Path to the output RDF file, if specified.
        validate (bool):
            Determines if data-schema validation checks will be performed. True by default.
        include_cells (bool):
            Determines if cell data will be included in the RDF output. True by default.
    Returns:
        An RDFlib graph object.
    """
    if not cas_schema:
        cas_schema = "base"
    base_linkml_schema = read_schema(cas_schema)

    data = get_instance_data(data)
    labelsets = [labelset['name'] for labelset in sorted(data["labelsets"], key=lambda x: x['rank'])]

    decorated_schema = decorate_linkml_ontology_schema(
        base_linkml_schema,
        ontology_namespace=ontology_namespace,
        ontology_iri=ontology_iri,
        labelsets=labelsets,
    )
    expanded_schema = expand_schema(
        config=None, yaml_obj=decorated_schema, value_set_names=["CellTypeEnum"]
    )

    # Prepare the data
    instance = populate_ids(
        data,
        ontology_namespace=ontology_namespace,
        ontology_id=ontology_namespace,
    )
    rdf_graph = dump_to_rdf(
        schema=expanded_schema,
        instance=instance,
        ontology_namespace=ontology_namespace,
        ontology_iri=ontology_iri,
        schema_name=get_schema_name(cas_schema),
        validate=validate,
        include_cells=include_cells,
        output_path=output_path,
    )

    return rdf_graph


def get_instance_data(data: Union[str, dict]):
    """
    Reads the data from the file or returns the data as is.
    Args:
        data: The data JSON file path or JSON object dictionary.
    Returns: The data as a dictionary.
    """
    if isinstance(data, Path):
        data = str(data)
    if isinstance(data, str):
        data = get_json_from_file(data)
        if data is None:
            raise ValueError("No such file: " + str(data))
    return data


def get_schema_name(cas_schema: str) -> str:
    """
    Returns the schema name based on the CAS schema name.
    Args:
        cas_schema (str): Name of the CAS schema.
    Returns:
        str: The schema name.
    """
    if isinstance(cas_schema, str):
        if cas_schema.lower() == "base" or cas_schema.endswith("general_schema.yaml"):
            return "base"
        elif cas_schema.lower() == "bican" or cas_schema.endswith("BICAN_schema.yaml"):
            return "bican"
        elif cas_schema.lower() == "cap" or cas_schema.endswith("CAP_schema.yaml"):
            return "cap"
    return ""



import dataclasses

from jsonasobj2 import JsonObj


def asdict(taxonomy):
    """
    Convert a taxonomy python object to a dictionary.
    :param taxonomy: taxonomy object
    :return: dictionary
    """
    taxon_dict = dataclasses.asdict(taxonomy)

    for annotation in taxon_dict["annotations"]:
        # Convert JsonObj to dict
        if annotation.get("author_annotation_fields") and isinstance(
                annotation["author_annotation_fields"], JsonObj):
            annotation["author_annotation_fields"] = annotation["author_annotation_fields"]._as_dict

        # Convert CellTypeEnum to string
        if annotation.get("cell_ontology_term_id"):
            if type(annotation["cell_ontology_term_id"]).__name__ == "CellTypeEnum":
                annot_obj = get_annotation_with_name(taxonomy, annotation["cell_label"],
                                                     annotation["labelset"])
                annotation["cell_ontology_term_id"] = str(annot_obj.cell_ontology_term_id.code.text)

    # Remove None values
    def remove_none_values(d):
        if not isinstance(d, dict):
            return d
        return {k: remove_none_values(v) for k, v in d.items() if v is not None}
    taxon_dict = remove_none_values(taxon_dict)

    return taxon_dict


def get_annotation_with_name(taxonomy, cell_label, labelset, compare_func=None):
    """
    Get an annotation with a given cell label and labelset in the given taxonomy.
    :param taxonomy: taxonomy object
    :param cell_label: cell label
    :param labelset: labelset
    :param compare_func: optional function to compare cell labels (such as `lambda x, y: x.endswith(y)`)
    :return: annotation object. None if not found
    """
    cell_label = str(cell_label).strip()
    for annotation in taxonomy.annotations:
        if annotation.labelset == labelset:
            if compare_func:
                if compare_func(annotation.cell_label, cell_label):
                    return annotation
            else:
                if annotation.cell_label == cell_label:
                    return annotation
    return None



import glob
import sys
import os
import warnings

from pathlib import Path
from typing import List
from ruamel.yaml import YAML

from cell_annotation_schema.file_utils import get_json_from_file
from cell_annotation_schema.ontology.schema import decorate_linkml_schema

from linkml.validator import validate


warnings.filterwarnings("always")


def validate_data(schema, test_path):
    """
    Validates all json files located in the test path with the given schema.
    Parameters:
        schema: json schema object
        test_path: path to the data files. If path is a folder, validates all json files inside. If path is a json file,
        validates it.
    Returns:
        'True' if all test files are valid, 'False' otherwise. Logs the validation errors if any.
    """
    # sv = get_validator(schema, schema_name)
    if os.path.isdir(test_path):
        test_files = glob.glob(pathname=test_path + "/*.json")
    else:
        if Path(test_path).suffix == ".json":
            test_files = [test_path]
        else:
            raise Exception("Test file extension not supported: {}".format(test_path))
    validation_status: List[bool] = []
    print("Found %s test files in %s" % (str(len(test_files)), test_path))
    for instance_file in test_files:
        i = get_json_from_file(instance_file)
        print("Testing: %s" % instance_file)
        report = validate(i, schema)
        if report.results:
            for result in report.results:
                print("Validation ERROR: " + result.message)
        validation_status.append(False if report.results else True)
    return False not in validation_status


def get_schema(file_path):
    """
    Reads the json schema from the given location
    """
    with open(file_path, "r") as fs:
        try:
            ryaml = YAML(typ="safe")
            return ryaml.load(fs)
        except Exception as e:
            raise Exception("Yaml read failed:" + file_path + " " + str(e))


def run_validator(path_to_schema_dir, schema_file, path_to_test_dir):
    """Tests all instances in a test_folder against a single schema.
    Assumes all schema files in single dir.
    Assumes all *.json files in the test_dir should validate against the schema.
       * path_to_schema_dir:  Absolute or relative path to schema dir
       * schema_file: schema file name
       * test_dir: path to test directory (absolute or local to schema dir)
    """
    # Getting script directory, schema directory and test directory
    script_folder = Path(os.path.dirname(os.path.realpath(__file__)))
    schema_dir = Path(os.path.dirname(path_to_schema_dir))
    test_path = os.path.join(script_folder, os.path.dirname(path_to_test_dir))
    if not os.path.exists(os.path.join(script_folder, schema_dir)):
        raise Exception("Please provide valid path_to_schema_dir")
    if not os.path.exists(test_path):
        raise Exception("Please provide valid path_to_test_dir")
    else:
        schema_file_path = os.path.join(script_folder, schema_dir, schema_file)
        schema = get_schema(schema_file_path)
        schema = decorate_linkml_schema(schema)

        result = validate_data(schema, test_path)
        if not result:
            raise Exception("Validation Failed")


if __name__ == "__main__":
    run_validator(
        path_to_schema_dir="../../build/", schema_file="general_schema.yaml", path_to_test_dir="../../examples/"
    )
    run_validator(
        path_to_schema_dir="../../build/", schema_file="BICAN_schema.yaml", path_to_test_dir="../../examples/BICAN_schema_specific_examples/"
    )
    run_validator(
        path_to_schema_dir="../../build/", schema_file="CAP_schema.yaml", path_to_test_dir="../../examples/CAP_schema_specific_files/"
        # Need to simplify names
    )



"""
Linkml data class generator (gen-python) is generating problematic code because of the subclass and id mechanism we use.

Details of the issue:
Taxonomy-annotations doesn't have a id field so it uses the first slot as id slot (which is labelset).
BicanTaxonomy-annotations (with range Bican Annotations) has a id field so it uses the cell_set_accession field as id slot and works fine.

When we use the gen-python command on the schema, BicanTaxonomy dataclass post_init normalisation code is generated correctly as follows:
```
_normalize_inlined_as_dict(slot_name="annotations", slot_type=BicanAnnotation, key_name="cell_set_accession", keyed=True)
```

But the Taxonomy dataclass post_init normalisation code is generated as follows (expectedly but not correctly):
```
_normalize_inlined_as_dict(slot_name="annotations", slot_type=Annotation, key_name="labelset", keyed=False)
```

And since the BicanTaxonomy post_init is calling the super.post_init, it is trying to normalise the Annotation dataclass
 with the labelset field as id field which is not correct.

Solution this code applies:

We read schema into memory and remove `annotations` slot from the Taxonomy class and add it to BicanTaxonomy
(as if it is not inherited but its own slot) so that the problematic normalisation code is not generated.
"""
import os
import json

from pathlib import Path
from typing import Union
from linkml import generators
from dacite import from_dict

from linkml_runtime.linkml_model import SchemaDefinition
from linkml_runtime.loaders import yaml_loader
from linkml_runtime.utils.compile_python import compile_python

from cell_annotation_schema.file_utils import read_schema
from cell_annotation_schema.ontology.schema import decorate_linkml_schema
from cell_annotation_schema.datamodel.cell_annotation_schema import Taxonomy
from cell_annotation_schema.datamodel.bican.cell_annotation_schema import BicanTaxonomy
from cell_annotation_schema.datamodel.cap.cell_annotation_schema import CapTaxonomy

SOURCE_DIR = Path(__file__).parent.parent


def generate_data_class(cas_schema: Union[str, dict], class_path: str):
    """
    Generate data class from CAS schema.

    Args:
        cas_schema: CAS schema path or dict representing it.
        class_path: Output class path.
    Returns:
        str: Data class string.
    """
    schema_def = read_schema(cas_schema)
    schema_dict = decorate_linkml_schema(schema_def)
    schema_def = yaml_loader.load(schema_dict, target_class=SchemaDefinition)
    gen = generators.PythonGenerator(schema_def)
    output = gen.serialize()
    with open(class_path, "w") as class_file:
        class_file.write(output)


def get_py_instance(instance_dict, schema_name, schema_def, root_class=None):
    """
    Returns a Python instance of the schema class from the given data instance.
    Args:
        instance_dict: The data instance dictionary.
        schema_name: The name of the schema to be used for RDF generation.
        schema_def: The schema definition object.
        root_class: The root class of the schema if this is not a core (base,cap or bican) schema.
    Returns:
        The Python instance of the schema class.
    """
    fix_author_annotation_fields(instance_dict)
    for annotation in instance_dict.get("annotations", []):
        # remove empty cell_ontology_term_id to prevent enum value failure
        if annotation.get("cell_ontology_term_id", None) == "":
            annotation.pop("cell_ontology_term_id", None)
            annotation.pop("cell_ontology_term", None)

    py_inst = None
    if isinstance(schema_name, str):
        if schema_name.lower() == "base":
            py_inst = from_dict(data_class=Taxonomy, data=instance_dict)
        elif schema_name.lower() == "bican":
            py_inst = from_dict(data_class=BicanTaxonomy, data=instance_dict)
        elif schema_name.lower() == "cap":
            py_inst = from_dict(data_class=CapTaxonomy, data=instance_dict)

    if not py_inst:
        # unknown schema, dynamically generate the python module and instantiate the class
        gen = generators.PythonGenerator(schema_def)
        output = gen.serialize()
        python_module = compile_python(output)
        py_target_class = getattr(python_module, root_class)
        py_inst = py_target_class(**instance_dict)

    for annotation in py_inst.annotations:
        # fix the author_annotation_fields in the json representation to be a string
        if annotation.author_annotation_fields and isinstance(annotation.author_annotation_fields, str):
            deserialised = json.loads(annotation.author_annotation_fields)
            annotation.author_annotation_fields = deserialised
    return py_inst


def fix_author_annotation_fields(json_obj):
    """
    Fix the author_annotation_fields in the json representation to be a string.
    Args:
        json_obj: json_obj dictionary.
    """
    for annotation in json_obj["annotations"]:
        if annotation.get("author_annotation_fields", None):
            annotation["author_annotation_fields"] = json.dumps(annotation["author_annotation_fields"])


def get_root_class(schema_name):
    """
    Returns the root class of the schema based on the schema name.
    Args:
        schema_name: The name of the schema.
    Returns: The root class of the schema.
    """
    root_class = None
    if schema_name.lower() == "base":
        root_class = "Taxonomy"
    elif schema_name.lower() == "bican":
        root_class = "BicanTaxonomy"
    elif schema_name.lower() == "cap":
        root_class = "CapTaxonomy"
    return root_class


if __name__ == "__main__":
    generate_data_class("base", os.path.join(SOURCE_DIR, "datamodel/cell_annotation_schema.py"))
    generate_data_class("bican", os.path.join(SOURCE_DIR, "datamodel/bican/cell_annotation_schema.py"))
    generate_data_class("cap", os.path.join(SOURCE_DIR, "datamodel/cap/cell_annotation_schema.py"))



import os
import json
import yaml
import warnings

from typing import Union
from urllib.request import urlopen
from pathlib import Path
from importlib import resources
from linkml_runtime.linkml_model import SchemaDefinition
from linkml_runtime.loaders import yaml_loader

from cell_annotation_schema import schemas


def get_cas_object(cas_json: Union[dict, str], schema_name: str):
    """
    Returns a CAS python object from the given JSON path or dictionary. Path can be a local file path or a
    web URL.
    :param cas_json: path/url to the JSON file or a dictionary
    :param schema_name: name of the schema to use. Should be one of the predefined schema names:
    `base`, `cap`, `bican`.
    :return: CAS python object
    """
    from cell_annotation_schema.generator.dataclassgen import get_py_instance

    if isinstance(cas_json, str):
        if is_web_url(cas_json):
            json_obj = get_json_from_url(cas_json)
        else:
            json_obj = get_json_from_file(cas_json)
    elif isinstance(cas_json, dict):
        json_obj = cas_json
    else:
        raise ValueError("Invalid input for cas_json (should be a dictionary or a path/url to a "
                         "JSON file).")

    taxonomy = get_py_instance(json_obj, schema_name, read_schema(schema_name))

    return taxonomy


def is_web_url(path):
    """
    Checks if the given path is a web URL.
    :param path: path reference to check
    :return: True if a web URL, False otherwise
    """
    return str(path).startswith("http://") or str(path).startswith("https://")


def get_json_from_url(url):
    """Loads JSOn from web URL."""
    import ssl
    ssl._create_default_https_context = ssl._create_unverified_context

    return json.loads(urlopen(url).read())


def get_json_from_file(filename):
    """Loads JSON from a file."""
    try:
        with open(filename, "r") as f:
            fc = f.read()
        return json.loads(fc)
    except FileNotFoundError:
        warnings.warn("File not found: " + filename)
    except IOError as exc:
        warnings.warn("I/O error while opening " + filename + ": " + str(exc))
    except json.JSONDecodeError as exc:
        warnings.warn("Failed to parse JSON in " + filename + ": " + str(exc))


def read_schema(schema: Union[str, dict]) -> SchemaDefinition:
    """
    Reads the given LinkML schema.
    Parameters:
        schema: The schema to read. When provided with a string, the input will first attempt to resolve to a predefined
         schema name (e.g., `base`, `cap`, `bican`). If it does not match any predefined schema name, it will be
         interpreted as a path, URL, or other loadable location. If the input is a dictionary, it should be compatible
         with `SchemaDefinition`. Otherwise, it should be an instance of `SchemaDefinition`.
    Returns: The SchemaDefinition object.
    """
    try:
        if isinstance(schema, str) and str(schema).lower() in get_cas_schema_names().keys():
            schema_name = get_cas_schema_names()[str(schema).lower()]
            schema_file = resources.files(schemas) / schema_name
            if os.path.exists(schema_file):
                # read from resources (schemas) package
                schema = yaml.safe_load(Path(schema_file).read_text())
            else:
                # read from build folder
                schema_dir = os.path.join(os.path.dirname(__file__), "../../build")
                schema = os.path.join(schema_dir, schema_name)
        if isinstance(schema, Path):
            schema = str(schema)
        if isinstance(schema, dict):
            schema = SchemaDefinition(**schema)
        elif isinstance(schema, str):
            schema = yaml_loader.load(schema, target_class=SchemaDefinition)

        if not isinstance(schema, SchemaDefinition):
            raise ValueError(f"Schema could not be loaded from {schema}")
    except ValueError as e:
        raise ValueError(f"Invalid schema: {schema}") from e
    return schema


def get_cas_schema_names() -> dict:
    """
    Returns the list of available CAS schema names.

    Returns:
        dict: The available CAS schema names.
    """
    return {
        "base": "general_schema.yaml",
        "cap": "CAP_schema.yaml",
        "bican": "BICAN_schema.yaml",
    }
